﻿using System;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Text.RegularExpressions;

using Platform.Win32Service;

namespace Platform.Modules.Implementation
{
    public class ModuleRunner : IWin32Service
    {
        #region Construction & Fields

        private const string RunAsServiceFlag = "-service";
        private const string RegisterServiceFlag = "-register";
        private const string UnregisterServiceFlag = "-unregister";
        private const string StartServiceFlag = "-start";
        private const string StopServiceFlag = "-stop";
        private const string HelpFlag = "-help";

        private ModuleHost _Host;

        /// <summary>
        /// Construct Module Runner
        /// </summary>
        /// <param name="host"></param>
        public ModuleRunner( ModuleHost host )
        {
            // store host
            _Host = host;
        }

        public string ServiceDisplayName { get; set; }
        public string ServiceDescription { get; set; }

        #endregion

        #region IWin32Service

        /// <summary>
        /// The Windows Service Name
        /// </summary>
        public string ServiceName => _Host.Name;

        /// <summary>
        /// Start Service
        /// </summary>
        /// <param name="startupArguments"></param>
        /// <param name="serviceStoppedCallback"></param>
        void IWin32Service.Start(string[] startupArguments, ServiceStoppedCallback serviceStoppedCallback)
        {
            _Host.Start();
        }

        /// <summary>
        /// Stop Service
        /// </summary>
        void IWin32Service.Stop()
        {
            _Host.Stop();
        }

        #endregion

        #region Tools

        private static string EscapeCommandLineArgument(string arg)
        {
            // http://stackoverflow.com/a/6040946/784387
            arg = Regex.Replace(arg, @"(\\*)" + "\"", @"$1$1\" + "\"");
            arg = "\"" + Regex.Replace(arg, @"(\\+)$", @"$1$1") + "\"";
            return arg;
        }

        #endregion

        #region Runner

        /// <summary>
        /// Start runner
        /// </summary>
        /// <param name="args"></param>
        public void Run( params string[] args )
        {
            try
            {
                if (args.Contains(RunAsServiceFlag))
                {
                    RunAsService(args);
                }
                else if (args.Contains(RegisterServiceFlag))
                {
                    RegisterService();
                }
                else if (args.Contains(UnregisterServiceFlag))
                {
                    UnregisterService();
                }
                else if (args.Contains(StartServiceFlag))
                {
                    StartService();
                }
                else if (args.Contains(StopServiceFlag))
                {
                    StopService();
                }
                else if (args.Contains(HelpFlag))
                {
                    DisplayHelp();
                }
                else
                {
                    RunInteractive(args);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine($"An error ocurred: {ex.Message}");
            }
        }

        private void RunAsService(string[] args)
        {
            var serviceHost = new Win32ServiceHost(this);
            serviceHost.Run();
        }

        /// <summary>
        /// Run as console application
        /// </summary>
        private void RunInteractive(string[] args)
        {
            // get assembly name
            var name = Assembly.GetEntryAssembly().GetName();

            // get log path
            var path = Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData);

            // start service
            Console.WriteLine("{0} v{1}", _Host.Name, name.Version);
            Console.WriteLine("Data Path={0}", path );
            Console.WriteLine("press enter to exit ...");
            Console.WriteLine();

            try
            {
                // start module host
                _Host.Start();
            }
            catch( Exception ex )
            {
                Console.WriteLine($"An error ocurred: {ex.Message}");
                Console.WriteLine(ex.StackTrace);
                Console.ReadLine();
                return;
            }

            // read key
            Console.ReadLine();

            // set shutdown
            Console.WriteLine();
            Console.WriteLine("Shutdown started ...");
            Console.WriteLine();

            // stop module host
            _Host.Stop();

            // say bye
            Console.WriteLine();
            Console.WriteLine("Bye !");
        }

        private void RegisterService()
        {
            // Environment.GetCommandLineArgs() includes the current DLL from a "dotnet my.dll --register-service" call, which is not passed to Main()
            var remainingArgs = Environment.GetCommandLineArgs()
                .Where(arg => arg != RegisterServiceFlag)
                .Select(EscapeCommandLineArgument)
                .ToList();
            remainingArgs.Add(RunAsServiceFlag);

            var host = Process.GetCurrentProcess().MainModule.FileName;

            if (!host.EndsWith("dotnet.exe", StringComparison.OrdinalIgnoreCase))
            {
                // For self-contained apps, skip the dll path
                remainingArgs.RemoveAt(0);
            }

            var fullServiceCommand = host + " " + string.Join(" ", remainingArgs);

            // Do not use LocalSystem in production.. but this is good for demos as LocalSystem will have access to some random git-clone path
            // Note that when the service is already registered and running, it will be reconfigured but not restarted
            new Win32ServiceManager()
                .CreateOrUpdateService(
                    ServiceName,
                    ServiceDisplayName,
                    ServiceDescription,
                    fullServiceCommand,
                    Win32ServiceCredentials.LocalSystem,
                    autoStart: true,
                    startImmediately: true,
                    errorSeverity: ErrorSeverity.Normal
                );

            Console.WriteLine($@"Successfully registered and started service ""{ServiceDisplayName}"" (""{ServiceDescription}"")");
        }

        private void UnregisterService()
        {
            new Win32ServiceManager()
                .DeleteService(ServiceName);

            Console.WriteLine($@"Successfully unregistered service ""{ServiceDisplayName}"" (""{ServiceDescription}"")");
        }

        private void StartService()
        {
            new Win32ServiceManager()
                .StartService(ServiceName);

            Console.WriteLine($@"Successfully started service ""{ServiceDisplayName}"" (""{ServiceDescription}"")");
        }

        private void StopService()
        {
            new Win32ServiceManager()
                .StopService(ServiceName);

            Console.WriteLine($@"Successfully stopped service ""{ServiceDisplayName}"" (""{ServiceDescription}"")");
        }

        private void DisplayHelp()
        {
            Console.WriteLine(ServiceDescription);
            Console.WriteLine();
            Console.WriteLine("Use one of the following options:");
            Console.WriteLine("  -register     Registers and starts this program as a windows service named \"" + ServiceDisplayName + "\"");
            Console.WriteLine("  -unregister   Removes the windows service");
            Console.WriteLine("  -start        Start the windows service");
            Console.WriteLine("  -stop         Stop the windows service");
        }

        #endregion
    }
}
